import numpy as np
from matplotlib.patches import Ellipse
import matplotlib.pyplot as plt


def make_fisher_matrix(params_dict, fisher_params, hpeak=0.0, obs='GS',
                        sigma=None, sigma_mod_frac=0.,
                        k_min=None, k_max=None,
                        z_min=None, z_max=None,
                        axis_PS=None, cosmo_key='CDM',
                        add_sigma_poisson=False):
    """
    Make Fisher matrix and its inverse from global signal or powerspectra

    Parameters
    ----------
    params_dict : dict
        Dictionary of parameter objects
    fisher_params : list
        List of parameter strings to use for Fisher matrix (these strings must be the keys to params_dict)
    hpeak : float
        TODO
    obs : str
        'GS' - global signal, 'PS' - power spectrum
    sigma : None,array
        TODO
    sigma_mod_frac : float
        Fraction of modelling error in PS e.g. 0.2 adds a 20% error on the PS in quadrature to the 21cmsense error
    k_min : None,float
        Minimum k to use for PS [1/Mpc]
    k_max : None,float
        Maximum k to use for PS [1/Mpc]
    z_min : None,float
        Minimum redshift to use for PS
    z_max : None,float
        Maximum redshift to use for PS
    axis_PS : None,int
        TODO
    cosmo_key : None,str
        TODO
    add_sigma_poisson : bool
        TODO

    Return
    -------
    Fisher matrix, Finv matrix
    """

    Fij_matrix = np.zeros((len(fisher_params), len(fisher_params)))

    for i,p1 in enumerate(fisher_params):

        if i == 0 and obs == 'PS':
            k_where = np.arange(len(params_dict[p1].PS_err[0]['k']))
            if k_min is not None and k_max is not None:  # k range in 1/Mpc
                k_where  = np.where((params_dict[p1].PS_err[0]['k'] <= k_max) & (params_dict[p1].PS_err[0]['k'] >= k_min))[0]

            z_where = np.arange(len(params_dict[p1].PS_z_HERA))
            if z_min is not None and z_max is not None:
                z_where  = np.where((params_dict[p1].PS_z_HERA <= z_max) & (params_dict[p1].PS_z_HERA >= z_min))[0]

            # Model error (e.g. 20%)
            sigma_mod = sigma_mod_frac * params_dict[p1].PS_fid[z_where][:,k_where]
            # if cosmo_key is None:
            # cosmo_key = params_dict[p1].deriv_PS.keys()[0]
            PS0 = params_dict[p1].deriv_PS[cosmo_key][z_where][:,k_where]

            # Poisson error
            if add_sigma_poisson:
                sigma_poisson = params_dict[p1].PS_err_Poisson[z_where][:,k_where]
            else:
                sigma_poisson = 0.

            # Fisher as a function of redshift or k?
            if axis_PS is not None:
                Fij_matrix = np.zeros((PS0.shape[axis_PS-1], len(fisher_params), len(fisher_params)))

        for j,p2 in enumerate(fisher_params):
            if obs == 'GS':
                if i == 0 and j == 0:
                    print('GS shape:',params_dict[p1].deriv_GS[cosmo_key].shape)

                Fij_matrix[i,j] = Fij(params_dict[p1].deriv_GS[cosmo_key],
                                      params_dict[p2].deriv_GS[cosmo_key],
                                      sigma_obs=1, sigma_mod=0.)
            elif obs == 'PS':
                if sigma is None:
                    sigma_PS = params_dict[p1].PS_sigma[z_where][:,k_where]
                else:
                    sigma_PS = sigma

                if i==0 and j==0:
                    print('PS shape:',params_dict[p1].deriv_PS[cosmo_key][z_where][:,k_where].shape)

                if axis_PS is not None:
                    Fij_matrix[:,i,j] = Fij(params_dict[p1].deriv_PS[cosmo_key][z_where][:,k_where],
                                          params_dict[p2].deriv_PS[cosmo_key][z_where][:,k_where],
                                          sigma_obs=sigma_PS, sigma_mod=sigma_mod, sigma_poisson=sigma_poisson, axis=axis_PS)
                else:
                    Fij_matrix[i,j] = Fij(params_dict[p1].deriv_PS[cosmo_key][z_where][:,k_where],
                                          params_dict[p2].deriv_PS[cosmo_key][z_where][:,k_where],
                                          sigma_obs=sigma_PS, sigma_mod=sigma_mod, sigma_poisson=sigma_poisson, axis=axis_PS)

    Finv = np.linalg.inv(Fij_matrix)
    return Fij_matrix, Finv


def Fij(dObs_dtheta_i, dObs_dtheta_j,
        sigma_obs=1., sigma_mod=0., sigma_poisson=0., axis=None):
    """
    Make fisher matrix elements

    Parameters
    ----------
    dObs_dtheta_i : array_like
        derivative wrt theta_i
    dObs_dtheta_j : array_like
        derivative wrt theta_j
    sigma_obs : array_like
        measurement uncertainties in the observations
    sigma_mod :
        modelling uncertainty
    axis : None or int or tuple of ints, optional
        Axis or axes along which a sum is performed. The default, axis=None,
        will sum all of the elements of the input array. If axis is negative
        it counts from the last to the first axis.

    Return
    ------
        F_ij fisher matrix element : float
    """
    sigma_sq = sigma_obs**2. + sigma_mod**2. + sigma_poisson**2.
    return np.nansum(dObs_dtheta_i * dObs_dtheta_j/sigma_sq, axis=axis)


def fisher_correlations(Fij_matrix, fisher_params, plot=True):
    """
    Fisher correlation matrix

    Parameters
    ----------
    Fij_matrix : array_like
        Fisher information matrix

    fisher_params: list
        ordered list of parameters in the fisher matrix

    plot: bool
        heatmap plot

    Return
    ------
        R_ij_fisher correlation matrix
    """

    R_ij_fisher = np.zeros((len(fisher_params), len(fisher_params)))
    for i,p1 in enumerate(fisher_params):
        for j,p2 in enumerate(fisher_params):
            R_ij_fisher[i,j] = Fij_matrix[i,j]/np.sqrt(Fij_matrix[i,i]*Fij_matrix[j,j])

    if plot:
        mask = np.triu(np.ones_like(R_ij_fisher, dtype=bool))
        import seaborn as sns
        sns.heatmap(R_ij_fisher, mask=mask,
                    cmap='RdBu',
                    xticklabels=fisher_params,yticklabels=fisher_params,
                    square=True, linewidths=.5, cbar_kws={"shrink": 1, 'label':'Correlation $r_{ij}$'})

    return R_ij_fisher


# TODO generalize
alpha_std = {1.:1.52, 2.:2.48, 3.:3.44}

def get_ellipse_params(i: int, j: int, cov: np.array):
    """
    Extract ellipse parameters from covariance matrix.
    Based on Coe 2009

    Parameters
    ----------
        i : int
            index of parameter 1

        j : int
            index of parameter 2

        cov : array_like
            covariance matrix

    Return
    ------
        ellipse a, b, angle in degrees
    """

    # equations 1-4 Coe 2009. returns in degrees
    def length(cov, sign=1):
        """
        Calculate length of the ellipse semi-major/semi-minor axes

        Aka the eigenvalues of the covariance matrix
        """
        return np.sqrt(0.5*(cov[i,i] + cov[j,j]) + sign*np.sqrt(0.25*(cov[i,i] - cov[j,j])**2. + cov[i,j]*cov[j,i]))

    def angle_deg(cov):
        """
        Calculate angle of ellipse in degrees (anti-clockwise from x axis)

        Gets the quadrant right!
        """
        return np.degrees(0.5*np.arctan2(2*cov[i,j],(cov[i,i] - cov[j,j])))

    a = length(cov, sign=1)
    b = length(cov, sign=-1)
    t = angle_deg(cov)

    return a, b, t


def plot_ellipse(ax, par1, par2, parameters, fiducial, cov,
                 resize_lims=True, positive_definite=[],
                 N_std=[1.,2.,3.], plot_rescale = 4.,
                 kwargs=[{'ls': '-'}],
                 color='tab:blue',
                 default_kwargs={'lw':0}):
    """
    Plot N-sigma ellipses, from Coe 2009.

    Parameters
    ----------
        ax : matpotlib axis
            axis upon which the ellipses will be drawn

        par1 : string
            parameter 1 name

        par2 : string
            parameter 2 name

        parameters : list
            list of parameter names

        fiducial : array_like(ndim,)
            fiducial values of parameters

        cov : array_like(ndim,ndim,)
            covariance matrix

        color : string
            color to plot ellipse with

        positive_definite : list of string
            convenience input, parameter names passed in this list
            will be cut off at 0 in plots.

    Returns
    -------
        list of float : sigma_x, sigma_y, sigma_xy

    """

    # Find parameters in list
    params = parameters
    pind = dict(zip(params, list(range(len(params)))))
    i = pind[par1]
    j = pind[par2]

    sigma_x  = np.sqrt(cov[i,i])
    sigma_y  = np.sqrt(cov[j,j])
    sigma_xy = cov[i,j]

    a, b, theta = get_ellipse_params(i, j, cov=cov)

    # Plot for each N sigma
    for nn, N in enumerate(N_std):
        # use defaults and then override with other kwargs
        kwargs_temp = default_kwargs.copy()
        if len(kwargs) > 1:
            kwargs_temp.update(kwargs[nn])
        else:
            kwargs_temp.update(kwargs[0])
        kwargs_n = kwargs_temp

        e = Ellipse(
            xy=(fiducial[i], fiducial[j]),
            width=a * 2 * alpha_std[N], height=b * 2 * alpha_std[N],
            angle=theta, zorder=0, facecolor=color, **kwargs_n)

        ax.add_artist(e)
        e.set_clip_box(ax.bbox)

    # Rescale the axes a bit
    if resize_lims:
        if par1 in positive_definite:
            ax.set_xlim(max(0.0, -plot_rescale*sigma_x),
                        fiducial[i]+plot_rescale*sigma_x)
        else:
            ax.set_xlim(fiducial[i] - plot_rescale * sigma_x,
                        fiducial[i] + plot_rescale * sigma_x)

        if par2 in positive_definite:
            ax.set_ylim(max(0.0, fiducial[j] - plot_rescale * sigma_y),
                        fiducial[j] + plot_rescale * sigma_y)
        else:
            ax.set_ylim(fiducial[j] - plot_rescale * sigma_y,
                        fiducial[j] + plot_rescale * sigma_y)

    return sigma_x, sigma_y, sigma_xy


def title_double_ellipses(axes, labels,
                           chain=None,
                           med=None, sigma=None,
                           title_fontsize=18, title_pad=58,
                           vspace=0.,
                           color='k'
                           ):
    """
    Plot title with parameter constraints from 2 covariance matrixes/chains

    Parameters
    ----------
        axes : matpotlib axess
            axes upon which the titles will be added

        labels : list(ndim,)
            list of parameter names

        chain : array_like(ndim,), optional
            MCMC chain of parameters

        med : array_like(ndim,), optional
            list of median values

        sigma : array_like(ndim,)
            list of sigmas

        color : string
            color to plot ellipse with


    Returns
    -------
        None

    """

    if chain is not None:
        l, med, u = np.percentile(chain, [16,50,84], axis=0)
        q_m, q_p  = med - l, u - med

    for i in range(len(labels)):
        if med[i] < 100:
            fmt = "{{0:{0}}}".format('.2f').format
        else:
            fmt = "{{0:{0}}}".format('.0f').format

        if chain is not None:
            CI = r"${{{0}}}_{{-{1}}}^{{+{2}}}$"
            CI = CI.format(fmt(med[i]), fmt(q_m[i]), fmt(q_p[i]))
        else:
            CI = r"${{{0}}} \pm {{{1}}}$"
            CI = CI.format(fmt(med[i]), fmt(sigma[i]))

        axes[i,i].set_title(f'{labels[i]}', fontsize=title_fontsize, pad=title_pad)
        axes[i,i].annotate(f'{CI}',
                           xy=(0.5,1.05+vspace), ha='center',
                           xycoords='axes fraction', color=color)

    return


def plot_triangle(params, fiducial, cov, fig=None, ax=None,
                   positive_definite=[],
                   labels=None,
                   resize_lims=True,
                   N_std=[1.,2.], plot_rescale = 4.,
                   ellipse_color='tab:blue',
                   ellipse_kwargs=[{},
                                  {'alpha':0.5}],
                   title_fontsize=20,
                   xlabel_kwargs={'labelpad': 5, 'fontsize':18},
                   ylabel_kwargs={'labelpad': 5, 'fontsize':18},
                   fig_kwargs={'figsize': (8, 8)},
                   plot1D_kwargs={'c':'black', 'lw':1}):
    """
    Make a triangle plot from a covariance matrix

    Based on https://github.com/xzackli/fishchips-public/blob/master/fishchips/util.py

    Parameters
    ----------
        params : list of strings
            List of parameter strings

        fiducial : array
            Numpy array consisting of where the centers of ellipses should be

        cov : numpy array
            Covariance matrix to plot

        fig : optional, matplotlib figure
            Pass this if you already have a figure

        ax : array containing matplotlib axes
            Pass this if you already have a set of matplotlib axes

        positive_definite: list
            List of parameter strings which are positive definite

        resize_lims : bool
            Resize ellipse limits to scale of the errors [default = True]

        N_std : list
            List of number of standard deviations to plot

        labels : list
            List of labels corresponding to each dimension of the covariance matrix

        ellipse_kwargs : dict
            Keyword arguments for passing to the 1-sigma Matplotlib Ellipse call. You
            can change this to change the color of your ellipses, for example.

        xlabel_kwargs : dict
            Keyword arguments which are passed to `ax.set_xlabel()`. You can change the
            color and font-size of the x-labels, for example. By default, it includes
            a little bit of label padding.

        ylabel_kwargs : dict
            Keyword arguments which are passed to `ax.set_ylabel()`. You can change the
            color and font-size of the y-labels, for example. By default, it includes
            a little bit of label padding.

        fig_kwargs : dict
            Keyword arguments which are passed to `figure`. E.g. figsize

        plot1D_kwargs : dict
            Keyword arguments which are passed to `plt.plot()` for 1D gauss plot

    Returns
    -------
        fig, ax
            matplotlib figure and axis array
    """

    nparams = len(params)

    if ax is None or fig is None:
        print('generating new axis')
        fig, ax = plt.subplots(nparams, nparams, **fig_kwargs)

    if labels is None:
        labels = [(r'$\mathrm{' + p.replace('_', r'\_') + r'}$')
                  for p in params]

    # stitch together axes to row=nparams-1 and col=0
    # and turn off non-edge
    for ii in range(nparams):
        for jj in range(nparams):
            if ii == jj:
                ax[jj, ii].get_yaxis().set_visible(False)
                if ii < nparams-1:
                    ax[jj, ii].get_xaxis().set_ticks([])

            if ax[jj, ii] is not None:
                if ii < jj:
                    if jj < nparams-1:
                        ax[jj, ii].set_xticklabels([])
                    if ii > 0:
                        ax[jj, ii].set_yticklabels([])

                    if jj > 0:
                        # stitch axis to the one above it
                        if ax[0, ii] is not None:
                            ax[jj, ii].get_shared_x_axes().join(ax[jj, ii], ax[0, ii])
                    elif ii < nparams-1:
                        if ax[jj, nparams-1] is not None:
                            ax[jj, ii].get_shared_y_axes().join(ax[jj, ii], ax[jj, nparams-1])

    # call plot_ellipse
    for ii in range(nparams):
        for jj in range(nparams):
            if ax[jj, ii] is not None:

                if ii < jj:
                    plot_ellipse(ax[jj, ii], params[ii],
                                 params[jj], params, fiducial, cov,
                                 positive_definite=positive_definite,
                                 N_std=N_std, plot_rescale=plot_rescale,
                                 resize_lims=resize_lims,
                                 color=ellipse_color,
                                 kwargs=ellipse_kwargs)
                    if jj == nparams-1:
                        ax[jj, ii].set_xlabel(labels[ii], **xlabel_kwargs)
                        ax[jj, ii].xaxis.set_major_locator(plt.MaxNLocator(5))
                        for tick in ax[jj, ii].get_xticklabels():
                            tick.set_rotation(45)
                    if ii == 0:
                        ax[jj, ii].set_ylabel(labels[jj], **ylabel_kwargs)

                elif ii == jj:
                    # plot a gaussian if we're on the diagonal
                    sig = np.sqrt(cov[ii, ii])
                    if params[ii] in positive_definite:
                        x = np.linspace(fiducial[ii], fiducial[ii] + plot_rescale * sig, 100)
                    else:
                        x = np.linspace(fiducial[ii] - plot_rescale*sig, fiducial[ii] + plot_rescale*sig, 100)

                    # Need to rescale if positive definite
                    rescale = 1.0
                    if params[ii] in positive_definite:
                        rescale = 2.0

                    gauss = rescale * np.exp(-(x-fiducial[ii])**2 / (2 * sig**2)) / (sig * np.sqrt(2*np.pi))
                    ax[jj, ii].plot(x, gauss, **plot1D_kwargs)
                    ax[jj, ii].set_title(f'{labels[ii]}$={fiducial[ii]:.2f} \pm {sig:.2f}$', fontsize=title_fontsize)
                    if ii == nparams-1:
                        ax[jj, ii].set_xlabel(labels[ii], **xlabel_kwargs)
                else:
                    ax[jj, ii].axis('off')

    fig.tight_layout()
    fig.subplots_adjust(hspace=0.05, wspace=0.05)

    return fig, ax



import py21cmfast as p21c
import os
import numpy as np
import glob
import scipy.optimize

import matplotlib.pyplot as plt

from .power_spectra import *
from .io import *


class Parameter(object):
    """Class for creating derivatives given 21cm parameters"""
    def __init__(self, param,
                HII_DIM=200, BOX_LEN=400,
                min_redshift=5.,
                n_chunks=24,
                k_PEAK_order=2.,
                output_dir=base_path+'examples/data/',
                PS_err_dir=base_path+'examples/data/21cmSense_noise/21cmSense_fid_EOS21/',
                Park19=None,
                k_HERA=True,
                cosmology='CDM',
                clobber=False,
                new=False,
                fid_only=False,
                vb=True):

        """

        Parameters
        ----------
        param : str
            Name of parameter (must be the same as 21cmFAST AstroParam)

        HII_DIM : int
            TODO

        BOX_LEN : int
            TODO

        Park19 : None, str
            Use Park+19 z bins when calculating power spectra
            https://ui.adsabs.harvard.edu/abs/2019MNRAS.484..933P/abstract
            'approx' = use our approximation of Park19 noise bins
            'real' = use noise from Jaehong

        k_HERA : bool
            If True (default), use HERA k bins for PS

        cosmology : str
            Label for cosmology, will be dict key for PS etc, default 'CDM'

        PS_err_dir : str
            Directory for errors, if making derivatives, assumes this exists

        clobber : bool
            If False - if paths for .npy file exist, load them

        new : bool
            If True, this is a new parameter -- let's make new GS and PS files from the lightcones

        vb : bool
            Verbose?
        """
        self.vb = vb

        self.param = param
        print('########### fisher set up for',self.param)

        self.k_PEAK_order = k_PEAK_order
        if self.param == 'k_PEAK':
            self.param_21cmfast = 'log10_k_PEAK'
            self.fid_i = 0
            if self.vb: print(f'    param = k_PEAK^-{self.k_PEAK_order}')
        else:
            self.param_21cmfast = self.param
            self.fid_i = 1

        self.param_label = self.param
        if self.param == 'L_X' or 'F' in self.param or self.param == 'M_TURN':
            self.param_label = 'log10 '+self.param

        self.output_dir = output_dir
        self.PS_err_dir = PS_err_dir

        self.HII_DIM = HII_DIM
        self.BOX_LEN = BOX_LEN
        self.min_redshift = min_redshift
        self.cosmology = cosmology

        self.lightcones = None

        self.fid_only = fid_only

        # Lightcone node redshifts (for global signal etc)
        self.redshifts      = None
        self.redshifts_file = f'{self.output_dir}redshifts.npy'
        if os.path.exists(self.redshifts_file):
            self.redshifts = np.load(self.redshifts_file, allow_pickle=True)
            if self.vb: print('    Loaded redshifts')

        # All lightcone redshifts
        self.lc_redshifts   = None
        self.lc_redshifts_file = f'{self.output_dir}lc_redshifts.npy'
        if os.path.exists(self.lc_redshifts_file):
            self.lc_redshifts = np.load(self.lc_redshifts_file, allow_pickle=True)
            if self.vb: print('    Loaded redshifts')

        self.T = None
        self.T_file = f'{self.output_dir}global_signal_dict_{self.param}.npy'
        if os.path.exists(self.T_file) and clobber is False:
            self.T = np.load(self.T_file, allow_pickle=True).item()
            if self.vb: print('    Loaded T(z) from',self.T_file)

        self.theta_params = None
        self.theta_file = f'{self.output_dir}params_dict_{self.param}.npy'
        if os.path.exists(self.theta_file) and clobber is False:
            self.theta_params = np.load(self.theta_file, allow_pickle=True).item()
            if self.vb: print('    Loaded param values from',self.theta_file)

        # lightcone chunks for PS
        self.n_chunks = n_chunks

        # Use HERA k bins to calculate all PS
        self.k_HERA = k_HERA

        # Power spectrum
        self.PS = None
        self.Park19 = Park19
        self.PS_suffix = ''
        if self.Park19 is not None:
            self.PS_suffix = '_Park19'

        # files to save
        self.PS_file = f'{self.output_dir}power_spectrum_dict_{self.param}{self.PS_suffix}.npy'
        self.PS_fid_file = f'{self.output_dir}power_spectrum_fid_21cmsense.npy'
        self.PS_z_HERA_file = f'{self.output_dir}PS_z_HERA{self.PS_suffix}.npy'
        self.PS_z_HERA = None
        self.deriv_GS = None
        self.deriv_PS = None

        # New fiducial
        if new and self.fid_only:
            print('    Fiducial only, get global signal and power spectra from the lightcones')

            # Load lightcones
            self.get_lightcones()

            # Make global signal and derivatives
            self.get_global_signal()

            # Load 21cmsense noise
            self.load_21cmsense(Park19=Park19)

            # Make power spectrum, load PS noise and make derivatives
            if self.k_HERA:
                k_min, k_max, n_psbins = self.get_HERA_k_bins_for_PS()
                self.get_power_spectra(k_min=k_min, k_max=k_max, n_psbins=n_psbins)
            else:
                self.get_power_spectra()

            self.make_PS_fid_HERA_grid()

        # New all parameters
        elif new:
            print('    New parameter, making new global signal and power spectra from the lightcones, make derivatives')
            # Load lightcones
            self.get_lightcones()

            # Make global signal and derivatives
            self.get_global_signal()
            self.derivative_global_signal()

            # Load 21cmsense noise
            self.load_21cmsense(Park19=Park19)

            # Make power spectrum, load PS noise and make derivatives
            if self.k_HERA:
                k_min, k_max, n_psbins = self.get_HERA_k_bins_for_PS()
                self.get_power_spectra(k_min=k_min, k_max=k_max, n_psbins=n_psbins)
            else:
                self.get_power_spectra()

            self.make_PS_fid_HERA_grid()
            self.derivative_power_spectrum()

        else:
            print('    Loading global signal and power spectra from saved files')

            # PS
            if os.path.exists(self.PS_file):
                self.PS = np.load(self.PS_file, allow_pickle=True).item()
                if self.vb: print('    Loaded PS from',self.PS_file)

            if os.path.exists(self.PS_fid_file):
                self.PS_fid = np.load(self.PS_fid_file, allow_pickle=True)
                if self.vb: print(f'    Loaded fiducial PS from {self.PS_fid_file}, shape:{self.PS_fid.shape}')

            # PS in HERA bins
            if os.path.exists(self.PS_z_HERA_file):
                self.PS_z_HERA = np.load(self.PS_z_HERA_file, allow_pickle=True)
                self.load_21cmsense(Park19=self.Park19)
                if self.vb: print('    Loaded PS_z_HERA from',self.PS_z_HERA_file,'shape=',self.PS_z_HERA.shape)

            # Derivatives
            if os.path.exists(self.T_file.replace('dict','deriv_dict')):
                self.deriv_GS = np.load(self.T_file.replace('dict','deriv_dict'), allow_pickle=True).item()
                if self.vb: print('    Loaded GS derivatives from',self.T_file.replace('dict','deriv_dict'))

            if os.path.exists(self.PS_file.replace('dict','deriv_dict')):
                self.deriv_PS = np.load(self.PS_file.replace('dict','deriv_dict'), allow_pickle=True).item()
                keys = list(self.deriv_PS.keys())
                if self.vb: print('    Loaded PS derivatives from',self.PS_file.replace('dict','deriv_dict'),'shape=',self.deriv_PS[keys[0]].shape)

        # Get fiducial Poisson noise
        try:
            self.load_Poisson_noise()
        except:
            print('Could not load Poisson noise, setting = None')
            self.PS_err_Poisson = None

        return



    def get_lightcones(self, regex=''):
        """
        Load lightcones and theta params
        """

        self.lightcones = []

        suffix = f'HIIDIM={self.HII_DIM}_BOXLEN={self.BOX_LEN}_fisher_*{regex}*{self.param}*'
        lightcone_filename = f'{self.output_dir}LightCone_z{self.min_redshift:.1f}_*{suffix}.h5'

        if self.vb: print(f'    Searching for lightcones with name {lightcone_filename}')
        lc_files = glob.glob(lightcone_filename)
        fid_filename = lightcone_filename.replace(self.param,'fid')
        fid_files = glob.glob(fid_filename)
        lc_files.extend(fid_files)

        # Don't get minihalo param lightcones if it's not the minihalo param!
        if 'MINI' not in self.param:
            lc_files = [f for f in lc_files if 'MINI' not in f]

        if self.vb: print(f'    Found {len(lc_files)} lightcones to load')

        for f in lc_files:
            self.lightcones.append(p21c.LightCone.read(f))
            print('    Loaded lightcones',f)

        self.k_fundamental, self.k_max, self.Nk = get_k_min_max(lightcone=self.lightcones[0], n_chunks=self.n_chunks)
        self.lc_redshifts = self.lightcones[0].lightcone_redshifts
        np.save(self.lc_redshifts_file, self.lc_redshifts, allow_pickle=True)

        return


    def get_global_signal(self, save=True, plot=False):
        """
        Get global signal and parameters and save to a file
        """
        self.T = {}
        self.theta_params = {}
        use_ETHOS = self.lightcones[0].flag_options.pystruct['USE_ETHOS']

        for lc in self.lightcones:
            if use_ETHOS:
                h_PEAK = np.round(lc.astro_params.pystruct['h_PEAK'],1)
                key = f'h_PEAK={h_PEAK:.1f}'
            else:
                key = self.cosmology

            if key not in self.T:
                self.T[key] = []
                self.theta_params[key] = []
            self.T[key].append(lc.global_brightness_temp)
            self.theta_params[key].append(lc.astro_params.pystruct[self.param_21cmfast])
            if self.redshifts is None:
                self.redshifts = lc.node_redshifts
                np.save(self.redshifts_file, self.redshifts, allow_pickle=True)

        for cosmo_key in self.T:
            self.T[cosmo_key] = np.array(self.T[cosmo_key])
            self.theta_params[cosmo_key] = np.array(self.theta_params[cosmo_key])

            if self.param == 'k_PEAK':
                self.theta_params[cosmo_key] = 1./self.theta_params[cosmo_key]**self.k_PEAK_order

            if self.param == 'L_X' or 'F' in self.param or self.param == 'M_TURN':
                self.theta_params[cosmo_key] = np.log10(self.theta_params[cosmo_key])  # make L_X, F log10

            # Sort increasing theta order
            self.T[cosmo_key] = self.T[cosmo_key][np.argsort(self.theta_params[cosmo_key])]
            self.theta_params[cosmo_key] = self.theta_params[cosmo_key][np.argsort(self.theta_params[cosmo_key])]

            # if len(self.theta_params[cosmo_key]) > 3:
            #     raise Attribute Error('Too many parameters',self.theta_params[cosmo_key])

            if plot:
                plt.plot(self.redshifts, self.T[cosmo_key].T)

        if save:
            np.save(self.T_file, self.T, allow_pickle=True)
            np.save(self.theta_file, self.theta_params, allow_pickle=True)
            if self.vb: print(f'    saved params to {self.theta_file}')
            if self.vb: print(f'    saved GS to {self.T_file}')

        return


    def derivative_global_signal(self, save=True, plot=True, ax=None):
        """
        Calculate global signal derivatives
        """

        if plot and ax == None:
            fig, ax = plt.subplots()

        self.deriv_GS = {}

        for cosmo_key in self.T:

            if 'h_PEAK=0.0' in cosmo_key:
                ls = 'dashed'
            else:
                ls = 'solid'

            if self.param == 'k_PEAK':
                deriv = np.zeros((len(self.theta_params[cosmo_key]),len(self.T[cosmo_key][0])))
                for j in range(len(self.theta_params[cosmo_key])):
                    if j > 0:
                        deriv[j] = (self.T[cosmo_key][j] - self.T[cosmo_key][0])/(self.theta_params[cosmo_key][j]-self.theta_params[cosmo_key][0])
                        if plot:
                            ax.plot(self.redshifts, deriv[j],
                                    lw=1, ls=ls,
                                    label='1/k_PEAK^%.1f < %.1e' % (self.k_PEAK_order, self.theta_params[cosmo_key][j]))

                self.deriv_GS[cosmo_key] = deriv[1]

            else:
                deriv = np.gradient(self.T[cosmo_key], self.theta_params[cosmo_key], axis=0)
                self.deriv_GS[cosmo_key] = deriv[1]

                labels = ['one-sided -','two-sided','one-sided +']
                if plot:
                    for dd,d in enumerate(deriv):
                        if len(deriv) <= 3:
                            ax.plot(self.redshifts, d, lw=1, ls=ls, label=labels[dd])
                        else:
                            ax.plot(self.redshifts, d, lw=1, ls=ls)


        if plot:
            ax.legend()
            ax.set_xlabel('Redshift')
            ax.set_ylabel(r'$\partial T_{21}/\partial \theta$')
            ax.set_title(self.param_label)

            fig.tight_layout()
            fig.savefig(self.output_dir+f'GS_deriv_{cosmo_key}_{self.param}.png', bbox_inches='tight')

        if save:
            GS_deriv_file = self.T_file.replace('dict','deriv_dict')
            np.save(GS_deriv_file, self.deriv_GS, allow_pickle=True)
            if self.vb: print(f'    saved GS derivatives to {GS_deriv_file}')

        return


    def get_HERA_k_bins_for_PS(self, plot=False):
        """
        Given k centers, find the bin edges and length to give to powerbox

        Because 21cmsense interpolates PS onto k grid based on HERA baselines etc,
        we must generate our PS and Poisson noise on that *same* k grid in order
        to get the errors for the Fisher.

        (Because if we used e.g. smaller k bins, our Poisson error would be too large)
        """
        k_centers = self.PS_err[0]['k'] * 0.7 # in Mpc^-1 (same in all frequency bins)

        def get_HERA_k_bins(k0, k_centers):
            """
            Given k centers and the first bin edge, find the bins
            """
            k_bins_iter = np.zeros(len(k_centers)+1)
            k_bins_iter[0] = k0

            # Predicted HERA k bins obtained iteratively
            i=0
            for i in range(len(k_centers)):
                log_k_bin_recover = 2*np.log(k_centers[i]) - np.log(k_bins_iter[i])
                k_bins_iter[i+1] = np.exp(log_k_bin_recover)

            return k_bins_iter

        def minimize_get_HERA_k_bins(k0, k_centers):
            """
            Find first k bin edge so that the k bins
            are equally spaced in log space
            """
            k_bins = get_HERA_k_bins(k0, k_centers)
            return np.std(np.diff(np.log10(k_bins)))


        # Find k0
        opt = scipy.optimize.minimize(minimize_get_HERA_k_bins, x0=k_centers[0]/2, args=k_centers)

        if plot:
            k0 = np.linspace(k_centers[0]/2,k_centers[0]*2,100)
            plt.plot(k0, [minimize_get_HERA_k_bins(k, k_centers) for k in k0])
            plt.axvline(opt['x'])

        # Get the bins
        k_bins_iter = get_HERA_k_bins(k0=opt['x'], k_centers=k_centers)
        k_min, k_max, n_psbins = k_bins_iter.min(), k_bins_iter.max(), len(k_bins_iter)

        # Predicted HERA k bins just from the min and max
        k_bins = np.logspace(np.log10(k_min),np.log10(k_max),n_psbins)

        # Calculate k centers, should be as close as possible to HERA k bins
        k = np.exp((np.log(k_bins[1:]) + np.log(k_bins[:-1])) / 2)

        assert (np.abs(k - k_centers) < 1e-5).all(), "ERROR: Could not recreate HERA k bins"

        if self.vb: print(f'    Found k_min, k_max, n_psbins to match 21cmsense noise')

        return k_min, k_max, n_psbins


    def get_power_spectra(self, n_psbins=50, k_min=None, k_max=None, save=True):
        """
        Make 21cm power spectra from redshift chunk list (bin edges)

        Parameters
        ----------
        n_psbins : int
            Number of k bins
        k_min : float, optional
            Minimum k value for PS [in 1/Mpc]
        k_max : float, optional
            Maximum k value for PS [in 1/Mpc]
        save : bool
            Save PS to file?
        """

        if self.Park19:
            # chunk_z_list_HERA = [27.408, 20.306, 16.0448, 13.204, 11.17485714,
            #                         9.653, 8.46933333, 7.5224, 6.74763636, 6.102,
            #                         5.55569231, 5.08742857]
            chunk_z_list_HERA = [27.15742, 22.97586, 19.66073, 16.98822, 14.80234,
                                12.99172, 11.4751, 10.19206, 9.09696, 8.15475,
                                7.33818, 6.62582, 6.0006]
        else:
            chunk_z_list_HERA = [27.4, 23.4828, 20.5152, 18.1892, 16.3171, 14.7778, 13.4898, 12.3962,
                                11.4561, 10.6393, 9.92308, 9.28986, 8.72603, 8.22078, 7.76543,
                                7.35294, 6.97753, 6.63441, 6.31959, 6.0297, 5.7619, 5.51376, 5.28319,
                                5.06838]#, 4.86777, 4.68]

        if self.lightcones is None:
            self.get_lightcones()
        chunk_indices_HERA = [np.argmin(np.abs(self.lc_redshifts - z_HERA)) for z_HERA in chunk_z_list_HERA][::-1]

        if self.vb: print(f'    Making powerspectra in {len(chunk_z_list_HERA)} redshift chunks and {n_psbins-1} k bins')

        self.PS = {}
        use_ETHOS = self.lightcones[0].flag_options.pystruct['USE_ETHOS']

        for lc in self.lightcones:
            if use_ETHOS:
                h_PEAK = np.round(lc.astro_params.pystruct['h_PEAK'],1)
                key = f'h_PEAK={h_PEAK:.1f}'
            else:
                key = self.cosmology

            theta  = lc.astro_params.pystruct[self.param_21cmfast]

            if self.param == 'k_PEAK':
                theta = 1./theta**self.k_PEAK_order

            if self.param == 'L_X' or 'F' in self.param or self.param == 'M_TURN':
                theta = np.log10(theta)  # make L_X, F log10

            if key not in self.PS:
                self.PS[key] = {} ##### TODO load PS nicely

            if self.vb: print(f'    Getting PS for {key}, {self.param}={theta}')

            # Make PS
            if k_min is None:
                k_min = self.k_fundamental
            if k_max is None:
                k_max = self.k_max

            if self.vb: print(f'        - Using k:{k_min}-{k_max}')

            self.PS_z_HERA, self.PS[key][f'{self.param}={theta}'] = powerspectra_chunks(lc,
                                                                 n_psbins=n_psbins,
                                                                 chunk_indices=chunk_indices_HERA,
                                                                 k_min=k_min,
                                                                 k_max=k_max,
                                                                 remove_nans=False)

            del lc

        if save:
            np.save(self.PS_file, self.PS, allow_pickle=True)
            if self.vb: print(f'    saved PS to {self.PS_file}')

            np.save(self.PS_z_HERA_file, self.PS_z_HERA, allow_pickle=True)
            if self.vb: print(f'    saved PS_z_HERA to {self.PS_z_HERA_file}')

        return


    def load_21cmsense(self, Park19=None):
        """
        Load 21cmsense errors from a given directory and save arrays to self

        Parameters
        ----------
        PS_err_dir : str, optional
            Directory where 21cmSense output is stored

        Park19: str, optional
            Use Park+19 z bins https://ui.adsabs.harvard.edu/abs/2019MNRAS.484..933P/abstract
            'approx' = use our approximation of Park19 noise bins
            'real' = use noise from Jaehong
        """

        # this can probably be removes
        # if Park19 == 'approx':
        #     self.PS_z_Park19 = sorted(np.array([23.3429, 17.9333, 14.4909, 12.1077, 10.36, 9.02353, 7.96842,
        #                                         7.11429, 6.4087, 5.816, 5.31111]))
        #     self.chunk_MHz = np.round(1420.4/(np.array(self.PS_z_Park19)+1), 0)
        # else:
        #     self.chunk_MHz = np.round(1420.4/(np.array(self.PS_z_HERA)+1), 0) # low redshift to high redshift

        if os.path.exists(self.PS_err_dir) is False:
            raise AttributeError(f'Path to 21cmsense errors: {self.PS_err_dir} does not exist/cannot be found - check your path')
        else:
            if self.vb: print(f'    Loading 21cmsense errors from {self.PS_err_dir}')

        PS_err   = []
        PS_sigma = []
        # PS_fid   = []
        if Park19 == 'real':
            Park19_noisefiles = np.array(glob.glob(self.PS_err_dir+f'../21cmSense_noise_Park19/TotalError_HERA331_PS_500Mpc_z*_1000hr.txt'))
            PS_z_Park19 = np.array([float(f.split('_z')[-1].split('_')[0]) for f in Park19_noisefiles])

            Park19_noisefiles = Park19_noisefiles[np.argsort(PS_z_Park19)] # low z to high z
            self.PS_z_Park19  = sorted(PS_z_Park19)

            # For each z bin
            for i, noise_file in enumerate(Park19_noisefiles):
                noise = np.genfromtxt(noise_file)

                PS_dict = {"k": noise[:,0], "delta":noise[:,1],
                            "err_mod":noise[:,1], "err_opt":noise[:,1], "err_pess":noise[:,1]}
                PS_err.append(PS_dict)
                PS_sigma.append(noise[:,1])
                # PS_fid.append(noise[:,1])

        else:
            err_files = sorted(glob.glob(self.PS_err_dir+f'Errlist_SplitCore_HERA350*'))
            frequencies = [f.split('_')[-1].split('.txt')[0] for f in err_files][::-1] # sort in decreasing frequency, increasing redshift
            k  = np.genfromtxt(self.PS_err_dir+f'klist_SplitCore_HERA350.drift_mod_{frequencies[-1]}.txt')
            for freq in frequencies:
                try:
                    delta = np.genfromtxt(self.PS_err_dir+f'P21model_SplitCore_HERA350.drift_mod_{freq}.txt')
                except:
                    delta = np.zeros_like(k)
                    if self.vb: print(f'    No delta file found, setting = nan')

                err_mod  = np.genfromtxt(self.PS_err_dir+f'Errlist_SplitCore_HERA350.drift_mod_{freq}.txt')
                try:
                    err_opt  = np.genfromtxt(self.PS_err_dir+f'Errlist_SplitCore_HERA350.drift_opt_{freq}.txt')
                    err_pess = np.genfromtxt(self.PS_err_dir+f'Errlist_SplitCore_HERA350.drift_pess_{freq}.txt')
                except:
                    err_opt = 0.
                    err_pess = 0.

                PS_dict = {"k": k, "delta":delta,
                            "err_mod":err_mod, "err_opt":err_opt, "err_pess":err_pess}
                PS_err.append(PS_dict)
                PS_sigma.append(err_mod)
                # PS_fid.append(delta)

        self.PS_err   = np.array(PS_err)
        self.PS_sigma = np.array(PS_sigma)
        # self.PS_fid   = np.array(PS_fid)

        return


    def make_PS_fid_HERA_grid(self):
        """
        Make fiducial PS in 21cmsense k bins [Mpc^-1]
        """
        if self.fid_only:
            fid_key = list(self.PS[self.cosmology].keys())[0]
        else:
            fid_key = sorted(list(self.PS[self.cosmology].keys()))[self.fid_i]
        print(f'    Fiducial: {fid_key}')

        # Make fiducial PS in 21cmsense k bins [Mpc^-1]
        ps_fid_all = self.PS[self.cosmology][fid_key]
        self.PS_fid = np.empty((len(self.PS_z_HERA),len(self.PS_err[0]['k'])))
        for i in range(len(self.PS_z_HERA)):
            k = ps_fid_all[i]['k']
            PS_interp = np.interp(self.PS_err[i]['k']*0.7, k, ps_fid_all[i]['delta'])
            self.PS_fid[i] = PS_interp

        np.save(self.PS_fid_file, self.PS_fid, allow_pickle=True)
        if self.vb: print(f'    saved fiducial PS to {self.PS_fid_file}')

        return


    def load_Poisson_noise(self):
        """
        Load Poisson noise from PS
        """
        if self.vb: print('    Loading Poisson noise for PS')

        PS_err_Poisson = []
        if self.fid_only:
            fid_key = list(self.PS[self.cosmology].keys())[0]
        else:
            fid_key = sorted(list(self.PS[self.cosmology].keys()))[self.fid_i]
        print(f'    Fiducial: {fid_key}')

        ps_fid_all = self.PS[self.cosmology][fid_key]
        for i in range(len(self.PS_z_HERA)):

            PS_err_Poisson_sim = ps_fid_all[i]['err_delta']

            k_sim = ps_fid_all[i]['k']
            k_err = self.PS_err[i]['k']*0.7 # h Mpc^-1 --> Mpc^-1

            if self.k_HERA is False: # this is a bad idea
                # interpolate onto 21cmsense k values
                PS_err_Poisson.append(np.interp(k_err, k_sim, PS_err_Poisson_sim))

            else: # this is right
                assert (np.abs(k_sim - k_err) < 1e-5).all(), "ERROR: simulated k bins do not match HERA k bins"
                PS_err_Poisson.append(PS_err_Poisson_sim)

        self.PS_err_Poisson = np.array(PS_err_Poisson)

        return


    def derivative_power_spectrum(self, save=True, plot=True, ax=None):
        """
        Calculate power spectrum derivatives

        Parameters
        ----------
        save : bool, optional
            Save PS derivative to file?

        plot : bool, optional
            Plot PS derivatives as a function of redshift

        ax : Union[None, plt.Axes]
            Matplotlib axes to plot on. If `None`, make a new figure
            with subplots
        """

        self.deriv_PS = {}

        for cosmo_key in sorted(self.T):

            if 'h_PEAK=0.0' in cosmo_key:
                ls = 'dashed'
            else:
                ls = 'solid'

            if plot:
                fig, ax = plt.subplots(4,int(np.round(len(self.PS_z_HERA)/4,0)),
                                        sharex=True, sharey=False, figsize=(15,9))
                ax = ax.ravel()
                fig.suptitle(cosmo_key+'  -  '+self.param)

            self.deriv_PS[cosmo_key] = np.zeros((len(self.PS_err),len(self.PS_err[0]['k'])))

            # For each z bin

            for i in range(len(self.PS_z_HERA)):

                # Get PS for each theta
                theta = []
                PS    = []
                for theta_key in self.PS[cosmo_key]:
                    theta.append(float(theta_key.split('=')[-1]))
                    PS.append(self.PS[cosmo_key][theta_key][i]['delta'])

                # # Add fiducial
                # if self.param != 'k_PEAK':
                #     theta_fid = self.theta_params[cosmo_key][1]
                # else:
                #     theta_fid = self.theta_params[cosmo_key][0]
                # theta.append(theta_fid)
                # PS.append(self.PS[cosmo_key][f'{self.param}={theta_fid}'][i]['delta'])

                k = self.PS[cosmo_key][theta_key][i]['k']

                # Sort PS in order of increasing theta
                # [x-h, x, x+h] for two-sided Derivatives (all astro params) where x is fid
                # [x, x+h1, x+h2] for asymmetric derivatives (1/kpeak)
                theta = np.array(theta)
                PS    = np.array(PS)[np.argsort(theta)]
                theta = theta[np.argsort(theta)]

                assert theta[-1] > theta[0], "thetas are not ordered correctly"

                if i==0:
                    print('theta =',theta)

                # Calculate derivative
                if self.param == 'k_PEAK':
                    deriv = np.zeros((len(theta),len(PS[0])))
                    for j in range(len(theta)):
                        if j > 0:
                            deriv[j] = (PS[j] - PS[0])/(theta[j]-theta[0])
                            if plot:
                                ax[i].semilogx(k, deriv[j],
                                        lw=1, ls=ls,
                                        label='1/k_PEAK^%.1f < %.1e' % (self.k_PEAK_order, theta[j]))
                else:
                    deriv = np.gradient(PS, theta, axis=0)
                    assert (deriv[1][~np.isnan(deriv[1])] == (PS[2][~np.isnan(deriv[1])]-PS[0][~np.isnan(deriv[1])])/(theta[2]-theta[0])).all(), 'two-sided derivative is wrong'

                if self.k_HERA:
                    self.deriv_PS[cosmo_key][i] = deriv[1]
                else:
                    # interpolate onto k for 21cmsense in 1/Mpc [21cmsense output is in h/Mpc]
                    self.deriv_PS[cosmo_key][i] = np.interp(self.PS_err[i]['k']*0.7, k, deriv[1])

                if plot:
                    try:
                        ax[i].set_title(f'z={self.PS_z_HERA[i]:.1f}')
                    except:
                        pass

                    if self.param != 'k_PEAK':
                        ax[i].semilogx(k, deriv.T, alpha=0.7)

                    ax[i].scatter(self.PS_err[i]['k']*0.7, self.deriv_PS[cosmo_key][i],
                                    c='k', s=5, ls='dashed', zorder=100)

                    ax[i].set_xlabel('k [$h$ Mpc$^{-1}$]')
                    ax[i].set_ylabel(r'$\partial \Delta^2_{21} (mK^2)/\partial \theta$')

            if plot:
                ax[0].legend()
                fig.tight_layout()
                fig.savefig(self.output_dir+f'PS_deriv_{self.param}_{cosmo_key}{self.PS_suffix}.png', bbox_inches='tight')

        if save:
            PS_deriv_file = self.PS_file.replace('dict','deriv_dict')
            np.save(PS_deriv_file, self.deriv_PS, allow_pickle=True)
            if self.vb: print(f'    saved PS derivatives to {PS_deriv_file}')

        return



# useful functions
import configparser
import os

code_path = os.path.abspath(os.path.dirname(__file__))
base_path = code_path.strip(code_path.split('/')[-1])

def read_config_params(item):
    """
    Read ints and booleans from config files
    Use for user_params and flag_options only

    Parameters
    ----------
    item : str
        config dictionary item as a string

    Return
    ------
    config dictionary item as an int, bool or str

    """
    try:
        return int(item)
    except:
        if item == 'True':
            return True
        if item == 'False':
            return False
        else:
            return item


def get_params_fid(config_file):
    """
    Load Fisher parameter names and fiducials from config file
    """
    config = configparser.ConfigParser(delimiters=':')
    config.optionxform = str
    config.read(config_file)

    astro_params_fid = dict(config.items('astro_params'))
    astro_params_fid = {key:float(astro_params_fid[key]) for key in astro_params_fid}

    astro_params_vary = config.get('vary','astro_params_vary').split('\n')
    astro_params_vary = list(filter(None, astro_params_vary))

    return astro_params_vary, astro_params_fid


astro_params_labels = {'ALPHA_ESC': r'$\alpha_\mathrm{esc}^{II}$',
                        'F_ESC10' : r'$\log_{10}f_\mathrm{esc,10}$',
                        'ALPHA_STAR' : r'$\alpha_\star^{II}$',
                        'F_STAR10' : r'$\log_{10}f_{\star,10}$',
                        'F_STAR7_MINI' : r'$\log_{10}f_\mathrm{\star,7}$',
                        'ALPHA_STAR_MINI' : r'$\alpha_\star^{III}$',
                        'F_ESC7_MINI' : r'$\log_{10}f_\mathrm{esc,7}$',
                        'L_X' : r'$\log_{10}\frac{L_X/{\dot{M}_\star}}{\mathrm{erg}\,\mathrm{s}^{-1}\, M_\odot^{-1}\,\mathrm{yr}}$',
                        'NU_X_THRESH' : r'$E_0$/eV',
                        'A_LW' : r'$A_\mathrm{LW}$',
                        'M_TURN': r'$\log_{10} (M_\mathrm{turn}/M_\odot)$',
                         't_STAR': r'$t_\star$'}



# -*- coding: utf-8 -*-

import glob
import os
import sys
import subprocess

from pkg_resources import DistributionNotFound, get_distribution
from unittest.mock import MagicMock

try:
    __version__ = get_distribution("21cmfish").version
except DistributionNotFound:
    __version__ = "unknown version"

sys.path.insert(0, os.path.abspath('../../'))

class Mock(MagicMock):
    @classmethod
    def __getattr__(cls, name):
        return MagicMock()

MOCK_MODULES = ['py21cmfast']
sys.modules.update((mod_name, Mock()) for mod_name in MOCK_MODULES)


# Convert the tutorials
for fn in glob.glob("_static/notebooks/*.ipynb"):
    name = os.path.splitext(os.path.split(fn)[1])[0]
    outfn = os.path.join("tutorials", name + ".rst")
    print("Building {0}...".format(name))
    subprocess.check_call(
        "jupyter nbconvert --template tutorials/tutorial_rst --to rst "
        + fn
        + " --output-dir tutorials",
        shell=True,
    )
    subprocess.check_call("python fix_internal_links.py " + outfn, shell=True)


extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.mathjax",
    "nbsphinx",
    'sphinx.ext.imgconverter'
]
templates_path = ["_templates"]
source_suffix = [".rst"]#,".ipynb"]
master_doc = "index"

project = "21cmfish"
copyright = "2021, Charlotte Mason & contributors"
version = __version__
release = __version__

# Readthedocs.
on_rtd = os.environ.get("READTHEDOCS", None) == "True"
if not on_rtd:
    import sphinx_rtd_theme

    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]

# #!/usr/bin/env python3
# # -*- coding: utf-8 -*-
# #
# # 21cmfish documentation build configuration file, created by
# # sphinx-quickstart on Thu Jun 28 12:35:56 2018.
# #
# # This file is execfile()d with the current directory set to its
# # containing dir.
# #
# # Note that not all possible configuration values are present in this
# # autogenerated file.
# #
# # All configuration values have a default; values that are commented out
# # serve to show the default.
#
# # If extensions (or modules to document with autodoc) are in another directory,
# # add these directories to sys.path here. If the directory is relative to the
# # documentation root, use os.path.abspath to make it absolute, like shown here.
# #
# # import os
# # import sys
# # sys.path.insert(0, os.path.abspath('.'))
#
#
# # -- General configuration ------------------------------------------------
#
# # If your documentation needs a minimal Sphinx version, state it here.
# #
# # needs_sphinx = '1.0'
#
# # Add any Sphinx extension module names here, as strings. They can be
# # extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# # ones.
# extensions = [
#     'sphinx.ext.autodoc',
#     # 'sphinx.ext.autosummary',
#     'sphinx.ext.githubpages',
#     'sphinx.ext.intersphinx',
#     'sphinx.ext.mathjax',
#     'sphinx.ext.viewcode',
#     'IPython.sphinxext.ipython_directive',
#     'IPython.sphinxext.ipython_console_highlighting',
#     'matplotlib.sphinxext.plot_directive',
#     'numpydoc',
#     # 'sphinx_copybutton',
#     'sphinx.ext.napoleon'
# ]
#
# # Configuration options for plot_directive. See:
# # https://github.com/matplotlib/matplotlib/blob/f3ed922d935751e08494e5fb5311d3050a3b637b/lib/matplotlib/sphinxext/plot_directive.py#L81
# plot_html_show_source_link = False
# plot_html_show_formats = False
#
# # Generate the API documentation when building
# autosummary_generate = True
# numpydoc_show_class_members = False
# napoleon_numpy_docstring = True
# napoleon_use_param = False
# napoleon_type_aliases = False
#
# # Add any paths that contain templates here, relative to this directory.
# templates_path = ['_templates']
#
# # The suffix(es) of source filenames.
# # You can specify multiple suffix as a list of string:
# #
# # source_suffix = ['.rst', '.md']
# source_suffix = '.rst'
#
# # The master toctree document.
# master_doc = 'index'
#
# # General information about the project.
# project = '21cmfish'
# copyright = '2021, Charlotte Mason'
# author = 'Charlotte Mason'
#
# # The version info for the project you're documenting, acts as replacement for
# # |version| and |release|, also used in various other places throughout the
# # built documents.
# #
# import py21cmfish
# # The short X.Y version.
# version = py21cmfish.__version__
# # The full version, including alpha/beta/rc tags.
# release = py21cmfish.__version__
#
# # The language for content autogenerated by Sphinx. Refer to documentation
# # for a list of supported languages.
# #
# # This is also used if you do content translation via gettext catalogs.
# # Usually you set "language" from the command line for these cases.
# language = None
#
# # List of patterns, relative to source directory, that match files and
# # directories to ignore when looking for source files.
# # This patterns also effect to html_static_path and html_extra_path
# exclude_patterns = []
#
# # The name of the Pygments (syntax highlighting) style to use.
# # pygments_style = 'sphinx'
#
# # If true, `todo` and `todoList` produce output, else they produce nothing.
# todo_include_todos = True
#
#
# # -- Options for HTML output ----------------------------------------------
#
# # The theme to use for HTML and HTML Help pages.  See the documentation for
# # a list of builtin themes.
# #
# html_theme = 'sphinx_rtd_theme'
# import sphinx_rtd_theme
# html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
#
# # Theme options are theme-specific and customize the look and feel of a theme
# # further.  For a list of options available for each theme, see the
# # documentation.
# #
# # html_theme_options = {}
#
# # Add any paths that contain custom static files (such as style sheets) here,
# # relative to this directory. They are copied after the builtin static files,
# # so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']
#
# # Custom sidebar templates, must be a dictionary that maps document names
# # to template names.
# #
# # This is required for the alabaster theme
# # refs: http://alabaster.readthedocs.io/en/latest/installation.html#sidebars
# html_sidebars = {
#     '**': [
#         'relations.html',  # needs 'show_related': True theme option to display
#         'searchbox.html',
#     ]
# }
#
#
# # -- Options for HTMLHelp output ------------------------------------------
#
# # Output file base name for HTML help builder.
# htmlhelp_basename = 'py21cmfish'
#
#
# # -- Options for LaTeX output ---------------------------------------------
#
# latex_elements = {
#     # The paper size ('letterpaper' or 'a4paper').
#     #
#     # 'papersize': 'letterpaper',
#
#     # The font size ('10pt', '11pt' or '12pt').
#     #
#     # 'pointsize': '10pt',
#
#     # Additional stuff for the LaTeX preamble.
#     #
#     # 'preamble': '',
#
#     # Latex figure (float) alignment
#     #
#     # 'figure_align': 'htbp',
# }
#
# # Grouping the document tree into LaTeX files. List of tuples
# # (source start file, target name, title,
# #  author, documentclass [howto, manual, or own class]).
# latex_documents = [
#     (master_doc, '21cmfish.tex', '21cmfish Documentation',
#      'Contributors', 'manual'),
# ]
#
#
# # -- Options for manual page output ---------------------------------------
#
# # One entry per manual page. List of tuples
# # (source start file, name, description, authors, manual section).
# man_pages = [
#     (master_doc, '21cmfish', '21cmfish Documentation',
#      [author], 1)
# ]
#
#
# # -- Options for Texinfo output -------------------------------------------
#
# # Grouping the document tree into Texinfo files. List of tuples
# # (source start file, target name, title, author,
# #  dir menu entry, description, category)
# texinfo_documents = [
#     (master_doc, '21cmfish', '21cmfish Documentation',
#      author, '21cmfish', 'A python package for doing Fisher matrix analyses with 21cmFAST',
#      'Miscellaneous'),
# ]
#
#
#
#
# # Example configuration for intersphinx: refer to the Python standard library.
# intersphinx_mapping = {
#     'python': ('https://docs.python.org/3/', None),
#     'numpy': ('https://numpy.org/doc/stable/', None),
#     'scipy': ('https://docs.scipy.org/doc/scipy/reference/', None),
#     'pandas': ('https://pandas.pydata.org/pandas-docs/stable', None),
#     'matplotlib': ('https://matplotlib.org/stable', None),
# }



# define functions to calculate PS, following py21cmmc
import numpy as np
from powerbox.tools import get_power

def get_k_min_max(lightcone, n_chunks=24):
    """
    Get the minimum and maximum k in 1/Mpc to calculate powerspectra for
    given size of box and number of chunks
    """

    BOX_LEN = lightcone.user_params.pystruct['BOX_LEN']
    HII_DIM = lightcone.user_params.pystruct['HII_DIM']

    k_fundamental = 2*np.pi/BOX_LEN*max(1,len(lightcone.lightcone_distances)/n_chunks/HII_DIM) #either kpar or kperp sets the min
    k_max         = k_fundamental * HII_DIM
    Nk            = np.floor(HII_DIM/1).astype(int)
    return k_fundamental, k_max, Nk


def compute_power(box,
                   length,
                   n_psbins,
                   log_bins=True,
                   k_min=None,
                   k_max=None,
                   ignore_kperp_zero=True,
                   ignore_kpar_zero=False,
                   ignore_k_zero=False):
    """
    Calculate power spectrum for a redshift chunk

    TODO

    Parameters
    ----------
    box :
        lightcone brightness_temp chunk

    length :
        TODO

    n_psbins : int
        number of k bins

    Returns
    ----------
        k : 1/Mpc
        delta : mK^2
        err_delta : mK^2
    """
    # Determine the weighting function required from ignoring k's.
    k_weights = np.ones(box.shape, dtype=int)
    n0 = k_weights.shape[0]
    n1 = k_weights.shape[-1]

    if ignore_kperp_zero:
        k_weights[n0 // 2, n0 // 2, :] = 0
    if ignore_kpar_zero:
        k_weights[:, :, n1 // 2] = 0
    if ignore_k_zero:
        k_weights[n0 // 2, n0 // 2, n1 // 2] = 0

    # Define k bins
    if k_min is None and k_max is None:
        bins = n_psbins
    else:
        if log_bins:
            bins = np.logspace(np.log10(k_min), np.log10(k_max), n_psbins)
        else:
            bins = np.linspace(k_min, k_max, n_psbins)

    res = get_power(
        box,
        boxlength=length,
        bins=bins,
        bin_ave=False,
        get_variance=True,
        log_bins=log_bins,
        k_weights=k_weights,
    )

    res = list(res)
    k = res[1]
    if log_bins:
        k = np.exp((np.log(k[1:]) + np.log(k[:-1])) / 2)
    else:
        k = (k[1:] + k[:-1]) / 2

    res[1] = k

    return res


def powerspectra(brightness_temp, n_psbins=50, nchunks=10,
                k_min=0.1, k_max=1.0, logk=True):
    """
    Make power spectra for given number of equally spaced chunks

    Output:
        k : 1/Mpc
        delta : mK^2
        err_delta : mK^2

    """
    data = []
    chunk_indices = list(range(0,brightness_temp.n_slices,round(brightness_temp.n_slices / nchunks),))

    if len(chunk_indices) > nchunks:
        chunk_indices = chunk_indices[:-1]
    chunk_indices.append(brightness_temp.n_slices)

    for i in range(nchunks):
        start = chunk_indices[i]
        end = chunk_indices[i + 1]
        chunklen = (end - start) * brightness_temp.cell_size

        power, k = compute_power(
            brightness_temp.brightness_temp[:, :, start:end],
            (BOX_LEN, BOX_LEN, chunklen),
            n_psbins,
            log_bins=logk,
        )
        data.append({"k": k, "delta": power * k ** 3 / (2 * np.pi ** 2)})
    return data


def powerspectra_np(brightness_temp, n_psbins=50, nchunks=10, k_min=0.1, k_max=1.0, logk=True):
    """
    Make power spectra for given number of equally spaced chunks

    JBM: same as powerspectra but for input in numpy format. Also outputs errors in delta
    """
    data = []

    mapshape = brightness_temp.shape
    chunk_indices = np.linspace(0, mapshape[-1], nchunks+1, endpoint=True,dtype=int)

    for i in range(nchunks):
        start = chunk_indices[i]
        end = chunk_indices[i + 1]
        chunklen = (end - start) * BOX_LEN/HII_DIM

        power, k, variance = compute_power(
            brightness_temp[:, :, start:end],
            (BOX_LEN, BOX_LEN, chunklen),
            n_psbins,
            log_bins=logk,
        )

        data.append({"k": k, "delta": power * k ** 3 / (2 * np.pi ** 2), "err_delta": np.sqrt(variance) * k ** 3 / (2 * np.pi ** 2)})
    return data


def powerspectra_chunks(lightcone, nchunks=10,
                        chunk_indices=None,
                        n_psbins=50,
                        k_min=0.1,
                        k_max=1.0,
                        logk=True,
                        model_uncertainty=0.15,
                        error_on_model=True,
                        ignore_kperp_zero=True,
                        ignore_kpar_zero=False,
                        ignore_k_zero=False,
                        remove_nans=True,
                        vb=False):

    """
    Make power spectra for given number of equally spaced redshift chunks OR list of redshift chunk lightcone indices

    Output:
        k : 1/Mpc
        delta : mK^2
        err_delta : mK^2

    TODO this isn't using k_min, k_max...
    """
    data = []

    # Create lightcone redshift chunks
    # If chunk indices not given, divide lightcone into nchunks equally spaced redshift chunks
    if chunk_indices is None:
        chunk_indices = list(range(0,lightcone.n_slices,round(lightcone.n_slices / nchunks),))

        if len(chunk_indices) > nchunks:
            chunk_indices = chunk_indices[:-1]

        chunk_indices.append(lightcone.n_slices)

    else:
        nchunks = len(chunk_indices) - 1

    chunk_redshift = np.zeros(nchunks)

    lc_redshifts = lightcone.lightcone_redshifts

    # Calculate PS in each redshift chunk
    for i in range(nchunks):
        if vb:
            print(f'Chunk {i}/{nchunks}...')
        start    = chunk_indices[i]
        end      = chunk_indices[i + 1]
        chunklen = (end - start) * lightcone.cell_size

        chunk_redshift[i] = np.median(lc_redshifts[start:end])

        if chunklen == 0:
            print(f'Chunk size = 0 for z = {lc_redshifts[start]}-{lc_redshifts[end]}')
        else:
            power, k, variance = compute_power(
                    lightcone.brightness_temp[:, :, start:end],
                    (lightcone.user_params.BOX_LEN, lightcone.user_params.BOX_LEN, chunklen),
                    n_psbins,
                    log_bins=logk,
                    k_min=k_min,
                    k_max=k_max,
                    ignore_kperp_zero=ignore_kperp_zero,
                    ignore_kpar_zero=ignore_kpar_zero,
                    ignore_k_zero=ignore_k_zero,)

            if remove_nans:
                power, k, variance = power[~np.isnan(power)], k[~np.isnan(power)], variance[~np.isnan(power)]
            else:
                variance[np.isnan(power)] = np.inf

            data.append({"k": k, "delta": power * k ** 3 / (2 * np.pi ** 2), "err_delta": np.sqrt(variance) * k ** 3 / (2 * np.pi ** 2)})

    return chunk_redshift, data
